{$D+,L+,Y+}

{****************************************}
{*                                      *}
{*           DSPRINT_.PAS 1.23          *}
{*          Borland Delphi unit         *}
{*   ( Version 1.X - 4.X compatible )   *}
{*                                      *}
{*      Copyright (c) Nikita Popov      *}
{*               1995-1998              *}
{*                                      *}
{*      Code revision : 30              *}
{*      Data revision : 10              *}
{*                                      *}
{****************************************}

{ Last update : 15-10-1998 }

{$I NIXLIBS.INC}

{$Q+}
{$S+}
{$R+}
{$X+}

unit DSprint; {- A single-component, single-step DataSet contents printing
  component }

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF WIN32}
  SysUtils, Messages, Classes, Graphics, Controls, Forms, Dialogs, Stdctrls,
  DB, Printers;

const
  defPPVoffs = 10; {- default vertical offset for the preview page in pixels }
  defPPHoffs = 10; {- default horizontal offset for the preview page in pixels }

const
  MMperInch = 25.4; {- millimeters per inch }

const
  DSPaligns : array[TAlignment] of word = ( DT_LEFT, DT_RIGHT, DT_CENTER );
  DSPwwraps : array[boolean] of word = ( 0, DT_WORDBREAK );
  DSPclrect : array[boolean] of word = ( 0, DT_CALCRECT );

{>>C 1.23: Added Delphi 3.X and 4.X field types support }
const
  dtsMemos : set of TFieldType =
  [ftBytes, ftVarBytes, ftBlob, ftMemo, ftGraphic{$IFDEF DELPHI3P}, ftFmtMemo,
   ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor{$ENDIF}{$IFDEF DELPHI4P}, ftFixedChar,
   ftWideString, ftLargeint, ftADT, ftArray, ftReference, ftDataSet
   {$ENDIF}];
  { Logical field types that are interpreted as "Memo fields" with specific
    printing features }

  dtsInvisible : set of TFieldType =
  [ ftUnknown, ftBytes, ftVarBytes, ftBlob{$IFDEF DELPHI3P}, ftFmtMemo,
   ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor{$ENDIF}{$IFDEF DELPHI4P}, ftFixedChar,
   ftWideString, ftLargeint, ftADT, ftArray, ftReference, ftDataSet
   {$ENDIF}];
  { Logical field types that are interpreted as "Invisible" or "Unprintable" }
{<<C 1.23}

type
  TPageHFtypesList = ( phfPrintDate, phfPrintTime, phfPageNo, phfHorzPieceNo,
    phfHorzPiecesTotal,
    {phfNumOfPages, phfCustomStrings,} phfBreakLines );
  TPageHFtype = set of TPageHFtypesList;

{>>A 1.23}
type
  TdspDataOption = (
    ddoHorzGridLines,  {- print horizontal grid lines }
    ddoVertGridLines   {- print vertical grid lines   }
);
  TdspDataOptions = set of TdspDataOption;
{<<A 1.23}

type
  TLineWrapperThreadingMode = ( lwrplSpace, lwrplCRLF );
  {- how "wrap" character will be interpreted: like "space" character or
     like a CR+LF characters' sequence }

type
  TDataSetPrint = class; {- forward declaration }

  TFieldCell = class {- data cell description class }
    constructor Create;
    destructor  Destroy; override;
  public
    Left   : integer;
    Width  : integer;
    Height : integer;
    Piece  : integer;
    Visible: boolean;
    UnPrn  : boolean;
{>>A 12-12-1997}
    Clipped : boolean;
    ClipLeft: integer;
{<<A 12-12-1997}
  end;

{>>A 1.23}
  TdspExtraGridRows = class(TPersistent)
    constructor Create(AOwner: TDataSetPrint);
    destructor  Destroy; override;
  private
    FRows : TStrings;
    FColCount: integer;
    FOwner: TDataSetPrint;
  protected
    function  GetCells(ACol, ARow: integer): string;
    procedure SetCells(ACol, ARow: integer; const Value: string);
    function  GetRowCount: integer;
    procedure SetRowCount(Value: integer);
    function  GetColCount: integer;
    procedure SetColCount(Value: integer);

    procedure Clear;
  public
    property Rows : TStrings read FRows;
    property Cells[ACol, ARow: integer]: string read GetCells write SetCells;
  published
    property RowCount: integer read GetRowCount write SetRowCount;
    property ColCount: integer read FColCount write SetColCount;
  end;
{<<A 1.23}

  TDataSetPrint = class(TComponent) {- master DSprint class }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  private
{-- Visible fields ----------------------------------------------------------}

    FDataSet         : TDataSet;
    FPageOrientation : TPrinterOrientation;

    FTitleFont       : TFont;
    FTitleWordWrap   : Boolean;
    fTitle           : TStrings;
    fTitleAlignment  : TAlignment;

    FDataHdrFont     : TFont;
    FDataHdrAlignment: TAlignment;
    FDataHdrWordWrap : boolean;

    FDataFont        : TFont;
    FWordWrapCells   : Boolean;

    fPageHeader      : TStrings;
    fPageHeaderOps   : TPageHFtype;
    fPHeaderFont     : TFont;
    fPHeaderAlignment: TAlignment;
    fPHeaderWordWrap : boolean;

    fPageFooter      : TStrings;
    fPageFooterOps   : TPageHFtype;
    fPFooterFont     : TFont;
    fPFooterAlignment: TAlignment;
    fPFooterWordWrap : boolean;

    FTailFont        : TFont;
    FTailWordWrap    : boolean;
    fTail            : TStrings;
    fTailAlignment   : TAlignment;

    FRestartData     : Boolean;
    FPrintVisibleOnly: Boolean;
    FPrintMemos      : Boolean;

    FLeftMargin    : integer;
    FTopMargin     : integer;
    FRightMargin   : integer;
    FBottomMargin  : integer;

    FVertLineWidth : integer;
    FHorzLineWidth : integer;
    FLockHVLwidth  : boolean;

{>>A 09-09-1996}
    FKerningCharacter : char;
{<<A 09-09-1996}

{>>A 10-10-1996}
    FDataTitleWrapper : char;
      {- character used to force new line in header }
    FDTWThreadingMode : TLineWrapperThreadingMode;
      {- how wrappers will be threaten during print process }

    FFitCellsByDataHdr : boolean;
{<<A 10-10-1996}

{>>A 12-12-1997}
    FOverlayPrinting: boolean;
    FOverlayWidthMM : byte;
{<<A 12-12-1997}

{>>A 02-03-1998}
    FNPagesToFit: integer;
{<<A 02-03-1998}

{>>A 18-03-1998}
    FPrintingDateTime: TDateTime;
{<<A 18-03-1998}
{>>A 1.23}
    FDataOptions: TdspDataOptions;
    FExtraGridRows: TdspExtraGridRows;
    FDataString   : string;
{<<A 1.23}

    FOnBeforePrint : TNotifyEvent;
    FOnAfterPrint  : TNotifyEvent;
    FOnBeginPage   : TNotifyEvent;
    FOnEndPage     : TNotifyEvent;
    FOnPrinterNewPage : TNotifyEvent;

{>>A 1.23}
    FOnBeforeExtraGridRows: TNotifyEvent;
    FOnAfterExtraGridRows : TNotifyEvent;
{<<A 1.23}
    FPrinterCanvas : TCanvas;

{-- Internal fields ----------------------------------------------------------}

    FXAspect        : real;
    FYAspect        : real;

    FPrintPreview  : Boolean;

    VLineWid : integer;
    HLineWid : integer;

    fFieldCells : TStringList;
    fBeginLeft : integer;
    fBeginTop  : integer;
    fEndRight  : integer;
    fEndBottom : integer;

    fHorzPieces: integer;

    DevPPI_X : integer;
    DevPPI_Y : integer;

    procedure SetDataSet(Value:TDataSet);
    procedure SetPageOrientation(Value:TPrinterOrientation);

    procedure SetDataHdrFont(Value:TFont);
    procedure SetDataFont(Value:TFont);
    procedure SetRestartData(Value:Boolean);

    procedure SetLeftMargin(Value : integer);
    procedure SetTopMargin(Value : integer);
    procedure SetRightMargin(Value : integer);
    procedure SetBottomMargin(Value : integer);

    procedure SetVertLineWidth(Value : integer);
    procedure SetHorzLineWidth(Value : integer);
    procedure SetLockHVLwidth(Value : boolean);

    procedure SetTitle(Value : TStrings);
    procedure SetTitleFont(Value:TFont);

    procedure SetPageHeader(Value : TStrings);
    procedure SetPHeaderFont(Value : TFont);

    procedure SetPageFooter(Value : TStrings);
    procedure SetPFooterFont(Value : TFont);

    procedure SetTail(Value : TStrings);
    procedure SetTailFont(Value : TFont);

    function  GetPrinterDC : HDC;
    function  GetReadyToPrint : boolean;

{>>A 09-09-1996}
    procedure SetKerningCharacter( Value : char );
{<<A 09-09-1996}

{>>A 10-10-1996}
    procedure SetDataTitleWrapper( Value : char );
    procedure SetDTWThreadingMode( Value : TLineWrapperThreadingMode );
    procedure SetFitCellsByDataHdr( Value : boolean );
{<<A 10-10-1996}

    procedure ClearFieldCells;

{>>A 02-03-1998}
    procedure SetNPagesToFit(Value: integer);
{<<A 02-03-1998}

  protected
    function  ReplaceWrappers( S : string; Mode : TLineWrapperThreadingMode ): string;
    function  GetMaxWrappedCharWidth( S : string ): integer;

    function  MM2PixelsVert( MMs : integer ) : integer;
    function  MM2PixelsHorz( MMs : integer ) : integer;

  public

{>>C 12-12-1997: added PieceNo parameter. If it's <> 0 then just specified
   piece must be printed }
    function  GenericPrint(const PieceNo: integer): boolean;
{<<C 12-12-1997}

    Function  Print : Boolean;
    Function  Preview : Boolean;

    function  CalculatePrinterProps: boolean;
    function  BuildFieldCells: boolean;
    function  GetCellHeight( S : string;
                             Rect : TRect;
                             WordWrap : boolean;
                             Alignment : TAlignment;
                             CV : TCanvas ) : integer;
{>>A 11-03-1998}
    function  GetCell(Index: integer): TFieldCell;
{>>A 11-03-1998}

{>>A 12-12-1997}
    function ScaleX(const RX: integer): integer;
    function ScaleY(const RY: integer): integer;
{<<A 12-12-1997}

    Property ReadyToPrint : boolean read GetReadyToPrint;
    Property PrinterDC : HDC read GetPrinterDC;
    Property PrintPreview : boolean read FPrintPreview write FPrintPreview default False;

    property HorzPieces: integer read FHorzPieces;
{>>A 11-03-1998}
    property Cells[Index: integer]: TFieldCell read GetCell;
{<<A 11-03-1998}
{>>A 18-03-1998}
    property PrintingDateTime: TDateTime read FPrintingDateTime;
{<<A 18-03-1998}
{- if assigned, print on this canvas }
    property PrinterCanvas: TCanvas read FPrinterCanvas write FPrinterCanvas;
  published
    Property DataSet : TDataSet read FDataSet write SetDataSet ;
    Property PageOrientation : TPrinterOrientation read FPageOrientation write SetPageOrientation ;

    Property RestartData : Boolean read FRestartData write SetRestartData ;
    Property PrintVisibleOnly : Boolean read FPrintVisibleOnly write FPrintVisibleOnly;
{    Property PrintMemos : Boolean read FPrintMemos write FPrintMemos default False; }

{- Data grid }
    Property DataHeaderFont : TFont read fDataHdrFont write SetDataHdrFont;
    Property DataHeaderAlignment : TAlignment read fDataHdrAlignment write fDataHdrAlignment;
    Property DataHeaderWordWrap : boolean read FDataHdrWordWrap write FDataHdrWordWrap;

    Property DataFont : TFont read FDataFont write SetDataFont ;
    Property WordWrapCells : Boolean read FWordWrapCells write FWordWrapCells ;

{- Report header ( title ) }
    Property rpTitle : TStrings read fTitle write SetTitle;
    Property rpTitleWordWrap : Boolean read FTitleWordWrap write FTitleWordWrap ;
    Property rpTitleFont : TFont read FTitleFont write SetTitleFont ;
    Property rpTitleAlignment : TAlignment read fTitleAlignment write fTitleAlignment;

{- Page header }
    Property rpPageHeader : TStrings read fPageHeader write SetPageHeader;
    Property rpPageHeaderOps : TPageHFtype read fPageHeaderOps write fPageHeaderOps;
    Property rpPageHeaderFont: TFont read fPHeaderFont write SetPHeaderFont;
    Property rpPageHeaderAlignment: TAlignment read fPHeaderAlignment write fPHeaderAlignment;
    Property rpPageHeaderWordWrap : boolean read fPHeaderWordWrap write fPHeaderWordWrap;

{- Page footer }
    Property rpPageFooter : TStrings read fPageFooter write SetPageFooter;
    Property rpPageFooterOps : TPageHFtype read fPageFooterOps write fPageFooterOps;
    Property rpPageFooterFont: TFont read fPFooterFont write SetPFooterFont;
    Property rpPageFooterAlignment: TAlignment read fPFooterAlignment write fPFooterAlignment;
    Property rpPageFooterWordWrap : boolean read fPFooterWordWrap write fPFooterWordWrap;

{- Report tail }
    Property rpTail : TStrings read fTail write SetTail;
    Property rpTailFont : TFont read fTailFont write SetTailFont;
    Property rpTailAlignment : TAlignment read fTailAlignment write fTailAlignment;
    Property rpTailWordWrap : boolean read fTailWordWrap write fTailWordWrap;

    Property MarginLeft_MM   : integer read FLeftMargin write SetLeftMargin;
    Property MarginTop_MM    : integer read FTopMargin write SetTopMargin;
    Property MarginRight_MM  : integer read FRightMargin write SetRightMargin;
    Property MarginBottom_MM : integer read FBottomMargin write SetBottomMargin;

    Property glnVLineWidth : integer read FVertLineWidth write SetVertLineWidth;
    Property glnHLineWidth : integer read FHorzLineWidth write SetHorzLineWidth;
    Property glnLockHorzVertWidths : boolean read FLockHVLwidth write SetLockHVLwidth;

    Property OnBeforePrint : TNotifyEvent read FOnBeforePrint write FOnBeforePrint ;
    Property OnAfterPrint : TNotifyEvent read FOnAfterPrint write FOnAfterPrint ;
    Property OnBeginPage : TNotifyEvent read FOnBeginPage write FOnBeginPage ;
    Property OnEndPage : TNotifyEvent read FOnEndPage write FOnEndPage ;
    Property OnPrinterNewPage : TNotifyEvent read FOnPrinterNewPage write FOnPrinterNewPage ;

{>>A 1.23}
    property OnBeforeExtraGridRows: TNotifyEvent read FOnBeforeExtraGridRows
      write FOnBeforeExtraGridRows;
    property OnAfterExtraGridRows : TNotifyEvent read FOnAfterExtraGridRows
      write FOnAfterExtraGridRows;
{<<A 1.23}

{>>A 09-09-1996}
    property KerningCharacter : char read FKerningCharacter write SetKerningCharacter;
    {- defines character that will be used as most widest alpha character
       when calculating cells' widths. Defaukt is "W" }
{<<A 09-09-1996}

{>>A 10-10-1996}
    property DataTitleWrapper : char read FDataTitleWrapper write SetDataTitleWrapper;
    {- defines character that will be treaten as title (data grid header) wrapper -
       character that forces to begin a new line ( like in InfoPower's wwDBgrid ).
       Default are "~" ( Tilda ) character. }

    property DTWThreading : TLineWrapperThreadingMode read fDTWThreadingMode
      write SetDTWThreadingMode;
    {- how wrappers will be threaten during print process }

    property FitCellsByHeader : boolean read FFitCellsByDataHdr
      write SetFitCellsByDataHdr;
    {- when true widths of the data columns will be calculated using widths
       of the columns' headers }
{<<A 10-10-1996}
{>>A 12-12-1997}
    property OverlayPrinting: boolean read FOverlayPrinting
      write FOverlayPrinting;
    {- when true columns which is not fitted in page in whole will be printed
       partially on its' starting page and on secondary page in "overlay"
       mode i.e. with some extra width to provide to user ability to attach
       pages from first to last as a chain with glue or staples. This can
       save much "blank" paper sheets if the DataSet has many columns with
       big width. }
    property OverlayWidthMM: byte read FOverlayWidthMM write FOverlayWidthMM;
{<<A 12-12-1997}
{>>A 02-03-1998}
    property NPagesToFit: integer read FNPagesToFit write SetNPagesToFit;
    {- defines a number of pages to "fit" printing to. When value of this
       property is greater than zero component will try to adjust widths
       of all columns which are addicted to be printed (i.e. they're Visible
       and Printable) to fit report output WIDTH to specified number of pages.
       In this case all columns' WIDTHes will be decreased proportionally for
       all printable columns. If this property value settled to zero no fitting
       actions will be done at all.

       NOTE: Specify too small number of pages in conjuction with big number
       of columns or it's great sizes may cause very ugly results
       so be careful. }
{<<A 02-03-1998}
{>>A 1.23}
    property rpDataOptions: TdspDataOptions read FDataOptions write FDataOptions;
    {- defines some options for data output. See the TdspDataOption type declaration }
    property ExtraGridRows: TdspExtraGridRows read FExtraGridRows;
{<<A 1.23}
  end;

procedure Register;

implementation {==============================================================}

uses
  PPreview,
  PProgrss,
  NXStrUtl,
  NixLibs,
  NxConsts;

{$IFDEF WIN32}
  {$R DSPRINT.D32}
  {$R DSPRINT.R32}
{$ELSE}
  {$R DSPRINT.D16}
  {$R DSPRINT.R16}
{$ENDIF}

const
  chTilda = '~';
  chTildaRpl : array[TLineWrapperThreadingMode] of char = ' '#13 ;

const
  wCrossMarkerWidth = 40;

procedure Register;
begin
  RegisterComponents(nxDBawarePalette, [TDataSetPrint]);
end;

{/// TFieldCell definitions //////////////////////////////////////////////////}

constructor TFieldCell.Create;
begin
  inherited Create;

end;

destructor TFieldCell.Destroy;
begin

  inherited Destroy;
end;

{/// TdspExtraGridRows ///////////////////////////////////////////////////////}
{>>A 1.23}
  
constructor TdspExtraGridRows.Create(AOwner: TDataSetPrint);
begin
  inherited Create;

  FOwner:= AOwner;
  FColCount:= 0;

  FRows:= TStringList.Create;
end;

destructor  TdspExtraGridRows.Destroy;
begin
  self.Clear;
  FRows.Free;

  inherited Destroy;
end;

function  TdspExtraGridRows.GetCells(ACol, ARow: integer): string;
begin
  if (GetColCount<=0) or (GetRowCount<=0) then
    Result:= EmptyStr
  else
    Result:= (FRows.Objects[ARow] as TStringList).Strings[ACol];
end;

procedure TdspExtraGridRows.SetCells(ACol, ARow: integer; const Value: string);
begin
  (FRows.Objects[ARow] as TStringList).Strings[ACol]:= Value;
end;

function  TdspExtraGridRows.GetRowCount: integer;
begin
  if FRows<>NIL then
    Result:= FRows.Count
  else
    Result:= 0;  
end;

procedure TdspExtraGridRows.SetRowCount(Value: integer);
var
  RowIdx, ColIdx: integer;
  Delta: integer;
  NewRow: TStringList;
begin
  if Value < 0 then
    Exit;

  Delta:= Value - FRows.Count;

  if Delta = 0 then
    Exit;

  if Delta < 0 then
  begin
    for RowIdx:= 1 to Delta do
      if FRows.Objects[FRows.Count-1]<>NIL then
      begin
        FRows.Objects[FRows.Count-1].Free;
        FRows.Delete(FRows.Count-1);
      end; { if }
  end
    else
  begin
    for RowIdx:= 1 to Delta do
    begin
      NewRow:= TStringList.Create;
      FRows.AddObject(EmptyStr, NewRow);

      for ColIdx:= 1 to FColCount do
        NewRow.Add(EmptyStr);
    end; { for }
  end; { if }
end;

function  TdspExtraGridRows.GetColCount: integer;
begin
  Result:= FColCount;
end;

procedure TdspExtraGridRows.SetColCount(Value: integer);
var
  ColIdx, RowIdx: integer;
  Delta: integer;
begin
  Delta:= Value - FColCount;

  if Delta = 0 then
    Exit;

  for RowIdx:= 0 to FRows.Count-1 do
    if Delta > 0 then
      for ColIdx:= 1 to Delta do
        (FRows.Objects[RowIdx] as TStringList).Add(EmptyStr)
    else
      for ColIdx:= 1 to Abs(Delta) do
        if (FRows.Objects[RowIdx] as TStringList).Count>0 then
           (FRows.Objects[RowIdx] as TStringList).Delete((FRows.Objects[RowIdx] as TSTringList).Count-1);

  FColCount:= Value;
end;

procedure TdspExtraGridRows.Clear;
begin
  while FRows.Count > 0 do
    if FRows.Objects[FRows.Count-1] <> NIL then
    begin
      FRows.Objects[FRows.Count-1].Free;
      FRows.Delete(FRows.Count-1);
    end; { if  }

  FColCount:= 0;
end;

{<<A 1.23}
{/// TDataSetPrint definition ////////////////////////////////////////////////}

constructor TDataSetPrint.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

  FFieldCells:= TStringList.Create;

  FDataHdrFont:= TFont.Create;
  FDataHdrFont.Name:= csDefPrintableFontName;
  FDataHdrFont.Size:= 11;
  FDataHdrFont.Style:= [fsBold];

  FdataHdrAlignment:= taCenter;

  FDataHdrWordWrap:= True;

  FDataFont:= TFont.Create;
  FDataFont.Name:= csDefPrintableFontName;
  FDataFont.Size:= 10;
  FDataFont.Style:= [];

  fTitle:= TStringList.Create;
  fTitleAlignment:= taCenter;
  fTitleWordWrap:= True;

  FTitleFont:= TFont.Create;
  FTitleFont.Name:= csDefPrintableFontName;
  FTitleFont.Size:= 10;
  FTitleFont.Style:= [fsBold];

{- Page header }
  fPageHeader:= TStringList.Create;
  fPHeaderAlignment:= taLeftJustify;
  fPHeaderWordWrap:= True;
  fPageHeaderOps:= [phfPrintDate, phfPrintTime];

  fPHeaderFont:= TFont.Create;
  fPHeaderFont.Name:= csDefPrintableFontName;
  fPHeaderFont.Size:= 10;
  fPHeaderFont.Style:= [];

{- Page footer }
  fPageFooter:= TStringList.Create;
  fPFooterAlignment:= taLeftJustify;
  fPFooterWordWrap:= True;
{>>C 17-03-1998}
  fPageFooterOps:= [phfPageNo, phfHorzPieceNo, phfHorzPiecesTotal];
{<<C 17-03-1998}

  fPFooterFont:= TFont.Create;
  fPFooterFont.Name:= csDefPrintableFontName;
  fPFooterFont.Size:= 10;
  fPFooterFont.Style:= [];

{- Report tail }
  FTailFont:= TFont.Create;
  FTailFont.Name:= csDefPrintableFontName;
  FTailFont.Size:= 10;
  FTailFont.Style:= [];

  fTail:= TStringList.Create;
  fTailAlignment:= taLeftJustify;
  fTailWordWrap:= True;

{- Other props }
  FVertLineWidth := 1;
  FHorzLineWidth := 1;
  FLockHVLwidth  := True;

  FXAspect:= 1;
  FYAspect:= 1;

  FHorzPieces:= 1;

  FPageOrientation:= poPortrait;

  FRestartData:= True;

  FPrintVisibleOnly:= True;

  FPrintMemos:= False;

  FPrintPreview:= False;
  FPrinterCanvas := nil;

{>>A 09-09-1996}
  FKerningCharacter:= 'W';
{<<A 09-09-1996}

{>>A 10-10-1996}
  FDTWThreadingMode:= lwrplCRLF;
  FDataTitleWrapper:= chTilda;

  FFitCellsByDataHdr:= False;
{<<A 10-10-1996}
{>>A 12-12-1997}
  FOverlayPrinting:= {True}False;
  FOverlayWidthMM:= 5;
{<<A 12-12-1997}
{>>A 02-03-1998}
  FNPagesToFit:= 0;
{<<A 02-03-1998}
{>>A 1.23}
  FDataOptions:= [ddoHorzGridLines, ddoVertGridLines];
  FExtraGridRows:= TdspExtraGridRows.Create(self);
{<<A 1.23}
end;

destructor TDataSetPrint.Destroy;
begin
  if fTail<>NIL then
    fTail.Free;

  if fTailFont<>NIL then
    fTailFont.Free;

  if fPageHeader<>NIL then
    fPageHeader.Free;

  if fPHeaderFont<>NIL then
    fPHeaderFont.Free;

  if fPageFooter<>NIL then
    fPageFooter.Free;

  if fPFooterFont<>NIL then
    fPFooterFont.Free;

  if fTitleFont<>NIL then
    FTitleFont.Free;

  if FDataFont<>NIL then
    FDataFont.Free;

  if FDataHdrFont<>NIL then
    FDataHdrFont.Free;

  if FFieldCells<>NIL then
  begin
    ClearFieldCells;
    FFieldCells.Free;
  end; { if }

  if FExtraGridRows<>NIL then
    FExtraGridRows.Free;

  inherited Destroy;
end;

{/// Properties //////////////////////////////////////////////////////////////}

procedure TDataSetPrint.SetDataSet(Value:TDataSet);
begin
  if FDataSet <> NIL then
    FDataSet:= NIL;
  if Value <> NIL then
    FDataSet:= Value;
end;

procedure TDataSetPrint.SetPageOrientation(Value:TPrinterOrientation);
begin
  FPageOrientation:= Value;
end;

procedure TDataSetPrint.SetDataHdrFont(Value:TFont);
begin
  fDataHdrFont.Assign( Value );
end;

procedure TDataSetPrint.SetDataFont(Value:TFont);
begin
  fDataFont.Assign( Value );
end;

procedure TDataSetPrint.SetRestartData(Value:Boolean);
begin
  if Value=fRestartData then Exit;

  fRestartData:= Value;
end;

function  TDataSetPrint.GetPrinterDC: HDC;
begin
  if Printer is TPrinter then Result:= Printer.Handle
                         else Result:= 0;
end;

function  TDataSetPrint.GetReadyToPrint : boolean;
begin
  Result:= False;

  Result:= ( FDataSet is TDataSet ) and (FDataSet.Active);
end;

{--- Set page margins --------------------------------------------------------}

procedure TDataSetPrint.SetLeftMargin(Value : integer);
begin
  if Value=FLeftMargin then Exit;

  if Value<0 then Exit;

  if Value>100 then Exit;

  fLeftMargin:= Value;
end;

procedure TDataSetPrint.SetTopMargin(Value : integer);
begin
  if Value=FTopMargin then Exit;

  if Value<0 then Exit;

  if Value>100 then Exit;

  fTopMargin:= Value;
end;

procedure TDataSetPrint.SetRightMargin(Value : integer);
begin
  if Value=FRightMargin then Exit;

  if Value<0 then Exit;

  if Value>100 then Exit;

  fRightMargin:= Value;
end;

procedure TDataSetPrint.SetBottomMargin(Value : integer);
begin
  if Value=FBottomMargin then Exit;

  if Value<0 then Exit;

  if Value>100 then Exit;

  fBottomMargin:= Value;
end;

procedure TDataSetPrint.SetVertLineWidth(Value : integer);
begin
  if Value=fVertLineWidth then Exit;

  if Value<1 then
  begin
    MessageDlg(LoadStr(SMinLineThiknessError), mtError, [mbOK], 0 );

    Exit;
  end; { if }

  if Value>16 then
  begin
    MessageDlg(LoadStr(SMaxLineThiknessError), mtError, [mbOK], 0 );

    Exit;
  end; { if }

  fVertLineWidth:= Value;

  if fLockHVLwidth then fHorzLineWidth:= Value;
end;

procedure TDataSetPrint.SetHorzLineWidth(Value : integer);
begin
  if Value=fHorzLineWidth then Exit;

  if Value<1 then
  begin
    MessageDlg(LoadStr(SMinLineThiknessError), mtError, [mbOK], 0 );

    Exit;
  end; { if }

  if Value>16 then
  begin
    MessageDlg(LoadStr(SMaxLineThiknessError), mtError, [mbOK], 0 );

    Exit;
  end; { if }

  fHorzLineWidth:= Value;

  if fLockHVLwidth then fVertLineWidth:= Value;
end;

procedure TDataSetPrint.SetLockHVLwidth(Value : boolean);
begin
  if Value=fLockHVLwidth then Exit;

  if Value then fHorzLineWidth:= fVertLineWidth;

  fLockHVLwidth:= Value;
end;

{--- Header, Footer, Page Header/Footer --------------------------------------}

procedure TDataSetPrint.SetTitle(Value : TStrings);
begin
  fTitle.Assign( Value );
end;

procedure TDataSetPrint.SetTitleFont(Value:TFont);
begin
  fTitleFont.Assign( Value );
end;

procedure TDataSetPrint.SetPageHeader(Value : TStrings);
begin
  fPageHeader.Assign( Value );
end;

procedure TDataSetPrint.SetPHeaderFont(Value:TFont);
begin
  fPHeaderFont.Assign( Value );
end;

procedure TDataSetPrint.SetPageFooter(Value : TStrings);
begin
  fPageFooter.Assign( Value );
end;

procedure TDataSetPrint.SetPFooterFont(Value:TFont);
begin
  fPFooterFont.Assign( Value );
end;

procedure TDataSetPrint.SetTailFont(Value:TFont);
begin
  fTailFont.Assign( Value );
end;

procedure TDataSetPrint.SetTail(Value : TStrings);
begin
  fTail.Assign( Value );
end;

{>>A 09-09-1996}
procedure TDataSetPrint.SetKerningCharacter( Value : char );
begin
  if Value=#0 then Exit;

  FKerningCharacter:= Value;
end;
{<<A 09-09-1996}

{>>A 10-10-1996}
procedure TDataSetPrint.SetDataTitleWrapper( Value : char );
begin
  if Value=FDataTitleWrapper then Exit;

  FDataTitleWrapper:= Value;
end;

procedure TDataSetPrint.SetDTWThreadingMode( Value : TLineWrapperThreadingMode );
begin
  if Value=FDTWThreadingMode then Exit;

  FDTWThreadingMode:= Value;
end;

procedure TDataSetPrint.SetFitCellsByDataHdr( Value : boolean );
begin
  if Value=FFitCellsByDataHdr then Exit;

  FFitCellsByDataHdr:= Value;
end;
{<<A 10-10-1996}

{>>A 02-03-1998}
procedure TDataSetPrint.SetNPagesToFit(Value: integer);
begin
  if Value=FNPagesToFit then Exit;

  if Value<0 then Exit;

  FNPagesToFit:= Value;
end;
{<<A 02-03-1998}

{>>A 11-03-1998}
function  TDataSetPrint.GetCell(Index: integer): TFieldCell;
begin
  Result:= fFieldCells.Objects[Index] as TFieldCell;
end;
{>>A 11-03-1998}

{/// Methods /////////////////////////////////////////////////////////////////}

{>>A 12-12-1997}
function TDataSetPrint.ScaleX(const RX: integer): integer;
begin
  if fXAspect=1 then Result:= RX else Result:= integer(round(RX * fXAspect));
end;

function TDataSetPrint.ScaleY(const RY: integer): integer;
begin
  if fYAspect=1 then Result:= RY else Result:= integer(round(RY * fYAspect));
end;
{<<A 12-12-1997}

{-----------------------------------------------------------------------------}

function  TDataSetPrint.ReplaceWrappers( S : string;
                          Mode : TLineWrapperThreadingMode ): string;
var
  B : byte;
begin
  for B:=1 to length(S) do
    if (S[B]=FDataTitleWrapper) and
        ( ( (B>1) and (S[B-1]<>FDataTitleWrapper) ) or
        ( B=1 ) ) then
       S[B]:= chTildaRpl[Mode];

  Result:= S;
end;

procedure TDataSetPrint.ClearFieldCells;
var
  Idx : integer;
begin
  for Idx:= FFieldCells.Count-1 downto 0 do
    (FFieldCells.Objects[Idx] as TFieldCell).Free;

  fFieldCells.Clear;

{>>A 1.23}
  if FExtraGridRows<>NIL then
    FExtraGridRows.ColCount:= 0;
{<<A 1.23}
end;

{-----------------------------------------------------------------------------}

function TDataSetPrint.CalculatePrinterProps: boolean;
begin
  Result:= False;

  Printer.Orientation:= fPageOrientation;

  fBeginLeft:= 0;
  fBeginTop:=  0;
  fEndRight:=  Printer.PageWidth;
  fEndBottom:= Printer.PageHeight;

  DevPPI_X:= GetDeviceCaps( Printer.Handle, LOGPIXELSX );
  DevPPI_Y:= GetDeviceCaps( Printer.Handle, LOGPIXELSY );

  fBeginLeft:= fBeginLeft + MM2PixelsHorz( fLeftMargin );
  fBeginTop:=  fBeginTop  + MM2PixelsVert( fTopMargin );
  fEndRight:=  fEndRight  - MM2PixelsHorz( fRightMargin );
  fEndBottom:= fEndBottom - MM2PixelsVert( fBottomMargin );

  if fPrintPreview then
  begin
    fXAspect:= 96 / DevPPI_X;
    fYAspect:= 96 / DevPPI_Y;
  end
    else
  begin
    fXAspect:= 1;
    fYAspect:= 1;
  end; { if }

  VLineWid:= fVertLineWidth;
  HLineWid:= fHorzLineWidth;

  if not fPrintPreview then
  begin
    VLineWid:=
      VLineWid * round(GetDeviceCaps(Printer.Handle, LOGPIXELSX )/96);

    if VLineWid<=0 then VLineWid:= 1;

    HLineWid:=
      HLineWid * round(GetDeviceCaps(Printer.Handle, LOGPIXELSY )/96);

    if HLineWid<=0 then HLineWid:= 1;
  end; { if }

  Result:= True;
end;

{-----------------------------------------------------------------------------}

function TDataSetPrint.BuildFieldCells: boolean;
var
  CellWid  : integer;
  CellHOfs : integer;

  PrevCWid : integer;
  PrevCHOfs: integer;

  FCell : TFieldCell;

  Idx : integer;
  SubIdx: integer;
{>>A 02-03-1998}
  OverallWidth: longint;
  ReqWidth: longint;
  CompressionRate: double;
  SubCompressionRate: double;
{<<A 02-03-1998}
begin
  Result:= False;

  ClearFieldCells;

  fHorzPieces:= 1;

  CellWid  := 0;
  CellHOfs := fBeginLeft;

  PrevCWid := 0;
  PrevCHOfs:= fBeginLeft;

  if fDataFont.Pitch=fpFixed then
  begin
    fDataFont.Name:= csDefPrintableFontName;
    fDataFont.Pitch:= fpVariable;
  end; { if }

  if fTitleFont.Pitch=fpFixed then
  begin
    fTitleFont.Name:= csDefPrintableFontName;
    fTitleFont.Pitch:= fpVariable;
  end; { if }

  Printer.Canvas.Font.Assign( fDataFont );
  Printer.Canvas.Font.Style:= fDataFont.Style;

{>>A 02-03-1998}
  OverallWidth:= 0;
{<<A 02-03-1998}

  for Idx:= 0 to DataSet.FieldCount-1 do
  begin {- building list of cells }
{>>C 10-10-1996}
    if FFitCellsByDataHdr then
    begin
      CellWid:= Printer.Canvas.TextWidth(FKerningCharacter)*
        GetMaxWrappedCharWidth( ReplaceWrappers(DataSet.Fields[Idx].DisplayLabel,
                                lwrplCRLF) );
    end
      else
    begin
      CellWid:= (DataSet.Fields[Idx].DisplayWidth+1) *
                 Printer.Canvas.TextWidth( FKerningCharacter );
    end; { if }

    if fPrintPreview then CellWid:= CellWid + 1
      else CellWid:= CellWid + fVertLineWidth*2;

{>>C 12-12-1997: removed try/except/end }
    FCell:= TFieldCell.Create;

    try
      fFieldCells.AddObject( DataSet.Fields[Idx].DisplayLabel,
                             FCell );
    except
      FCell.Free;

      raise;
    end; { try }
{<<C 12-12-1997 }

    FCell.UnPrn:= ( DataSet.Fields[Idx].DataType in dtsInvisible );

    FCell.Visible:= DataSet.Fields[Idx].Visible;

    if DataSet.Fields[Idx].DataType in dtsMemos then
      FCell.Visible:= FCell.Visible and fPrintMemos;

    if FCell.UnPrn then
      FCell.Visible:= False;

    if FCell.UnPrn then
      Continue;

    if fPrintVisibleOnly and (not FCell.Visible) then
      Continue;

    FCell.Clipped:= False;
    FCell.ClipLeft:= 0;

{>>A 02-03-1998}
    inc(OverallWidth, CellWid);
{<<A 02-03-1998}

{>>C 12-12-1997: added clipping code }
    if CellHOfs+CellWid > fEndRight then
    begin
      if FOverlayPrinting then
      begin
        FCell.ClipLeft:= (CellHOfs + CellWid) - fEndRight;
        FCell.Piece:= fHorzPieces;
        FCell.Left:=  CellHOfs;

        inc( fHorzPieces );

        PrevCHOfs:= CellHOfs;
        CellHOfs:= fBeginLeft;
      end
        else
      begin
        if FCell.Width>(fEndRight - fBeginLeft) then
        begin
          FCell.Piece:= fHorzPieces;
          inc( fHorzPieces );
        end
          else
        begin
          inc( fHorzPieces );
          FCell.Piece:= fHorzPieces;
        end; { if }

        PrevCHOfs:= CellHOfs;
        CellHOfs:= fBeginLeft;
        FCell.Left:=  CellHOfs;

        inc( CellHOfs, CellWid + VLineWid );
      end; { if }
    end
      else
    begin
      FCell.Piece:= fHorzPieces;
      FCell.Left:=  CellHOfs;

      PrevCHOfs:= CellHOfs;

      inc( CellHOfs, CellWid + VLineWid );
    end; { if }

    FCell.Width:= CellWid;
{<<C 12-12-1997 }

    PrevCWid:= CellWid;
  end; { for }

{- now the list of cells is completed and there we must apply "fit" action }

  if (FNPagesToFit > 0) and (fHorzPieces > FNPagesToFit) then
  begin
  {- We're cannot increate number of horz. pieces because already reached
     a limit of them. Trying to proportionally decrease widhes of all
     available cells including current to fit into specified number of
     horz. pieces. }

    ReqWidth:= (fEndRight - fBeginLeft - VLineWid) * FNPagesToFit;
    {- calculate required overall width }

    CompressionRate:= ReqWidth / OverallWidth;
    {- calculate the ratio to shrink all printable cells' widthes by }

    for Idx:= 0 to fFieldCells.Count-1 do
    if Cells[Idx].Visible then
      Cells[Idx].Width:= trunc(Cells[Idx].Width * CompressionRate) - VLineWid;

{ Now re-toss cells on pieces because they widthes has changes }
    CellHOfs:= fBeginLeft;
    PrevCHOfs:= CellHOfs;
    fHorzPieces:= 1;

    for Idx:= 0 to fFieldCells.Count-1 do
    if Cells[Idx].Visible then
    begin
      FCell:= Cells[Idx];

      if CellHOfs+FCell.Width > fEndRight then
      begin
        if FOverlayPrinting then
        begin
        end
          else
        begin
          if FCell.Width>(fEndRight - fBeginLeft) then
          begin
            FCell.Width:= (fEndRight - fBeginLeft - VLineWid);
          end;

          inc( fHorzPieces );
          FCell.Piece:= fHorzPieces;

          PrevCHOfs:= CellHOfs;
          CellHOfs:= fBeginLeft;
          FCell.Left:=  CellHOfs;

          inc( CellHOfs, CellWid + VLineWid );
        end; { if }
      end
        else
      begin
        FCell.Piece:= fHorzPieces;
        FCell.Left:=  CellHOfs;

        PrevCHOfs:= CellHOfs;

        inc( CellHOfs, FCell.Width + VLineWid );
      end; { if }
    end; { for if }

{ Now re-scale cells inside each piece if they're not fit into printable width }
    ReqWidth:= fEndRight-fBeginLeft;

    for Idx:= 1 to fHorzPieces do
    begin
      CellHOfs:= fBeginLeft;
      PrevCHOfs:= CellHOfs;
      OverallWidth:= 0;

      for SubIdx:= 0 to fFieldCells.Count-1 do
      if Cells[SubIdx].Visible and (Cells[SubIdx].Piece = Idx) then
      begin
        inc(OverallWidth, Cells[SubIdx].Width + VLineWid);
      end; { for }

      if OverallWidth > ReqWidth then
      begin
        CompressionRate:= ReqWidth / OverallWidth;
        {- calculate the ratio to shrink printable cells' widthes by }

        for SubIdx:= 0 to fFieldCells.Count-1 do
        if Cells[SubIdx].Visible and (Cells[SubIdx].Piece = Idx) then
        begin
          Cells[SubIdx].Left:= CellHOfs;

          Cells[SubIdx].Width:= trunc(Cells[SubIdx].Width * CompressionRate) -
            VLineWid;

          PrevCHOfs:= CellHOfs;

          inc(CellHOfs, Cells[SubIdx].Width + VLineWid);
        end; { for }
      end; { if }
    end; { for }
  end; { if }

{>>A 1.23}
  FExtraGridRows.ColCount:= FFieldCells.Count;
{<<A 1.23}    

  Result:= True;
end;

{-----------------------------------------------------------------------------}

function LgLeftPad( S : string ): string;
begin
  if S<>NullString then Result:= ' '+S else Result:= S;
end;

function TDataSetPrint.GetCellHeight;
var
  TMS : string;
  Idx : byte;
  C   : array[0..255] of char;
begin
{>>C 10-10-1996 }
  Result:= DrawText( CV.Handle,
                     StrPCopy( C, S ), -1,
                     Rect,
                     DT_CALCRECT or
                     DSPwwraps[WordWrap]
                     or DSPaligns[Alignment]
                     or DT_NOPREFIX
                     or DT_EXPANDTABS);

  if Result>=0 then Result:= Result + fHorzLineWidth;
end;

function TDataSetPrint.GetMaxWrappedCharWidth( S : string ): integer;
var
  WrPos : integer;
  WrLen : integer;
begin
  WrPos:= pos( #13, S );

  if WrPos<=0 then
  begin
    if trim(S)=NullString then S:= FKerningCharacter;
    Result:= length( S );

    Exit;
  end; { if }

  Result:= 0;
  while (length(S)>0) and (WrPos>0) do
  begin
    WrLen:= length( copy( S, 1, WrPos-1 ) );
    if WrLen>Result then Result:= WrLen;

    delete( S, 1, WrPos );
    WrPos:= pos( #13, S );

    if (WrPos<=0) and (length(S)>1) then WrPos:= length(S);
  end; { while }
end;

{/// Print methods ///////////////////////////////////////////////////////////}

Function TDataSetPrint.GenericPrint(const PieceNo: integer) : Boolean;
var
  PrnCV       : TCanvas;
  CurrPiece   : integer;
  DStateBKM   : TBookMark;
  BKM         : TBookMark;
  NewBKM      : TBookMark;
  PVoffset    : integer;
  CHt         : integer;
  GridHdrHt   : integer;

  CurrPageNo  : word;
  TotalPages  : word;

  LGDPageStart : integer; {- logical page start for data }
  LGDpageEnd   : integer; {- logical page end for data }

  LGSPageStart : integer; {- logical page start for service fields }
  LGSPageEnd   : integer; {- logical page end for service fields }

  LGDlastVofs  : integer;

  C : array[0..255] of char;
  ProgressCnt : longint;
  Cancelled   : boolean;

  HasErrors : boolean;
  IsLastPage: boolean;

  {---------------------------------------------------------------------------}
  procedure BeginPage;
  begin
    if fPrintPreview then
    begin
      PVoffset:= PVoffset+10;
    end
    else if Assigned(FPrinterCanvas) then
    begin
      if Assigned(FOnPrinterNewPage) then
      begin
        FOnPrinterNewPage(self);
        if Assigned(PrinterCanvas) then
        begin
          PrnCV := PrinterCanvas;
        end
      end;
    end
    else
    begin
      Printer.NewPage;
    end; { if }
  end;

  {--- print page header -----------------------------------------------------}

  function PrintPageHeader( Piece : integer ): boolean;
  var
    PageHdrWid: integer;
    OneLineHt : integer;
    LineIdx   : integer;
    HdrLine   : string;
    VOffset   : integer;
    ARect     : TRect;

  begin
    Result:= False;

    if Piece=1 then
    begin
      inc( LGSpageStart, round(10 * FYAspect) );
      if assigned( fOnBeginPage ) then fOnBeginPage( self );
    end; { if }

    VOffset:= LGSpageStart;

    PageHdrWid:= fEndRight-fBeginLeft;

    PrnCV.Font:= fPHeaderFont;
    PrnCV.Font.Color:= clBlack;
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Width:= HLineWid;
    PrnCV.Pen.Color:= clBlack;

    HdrLine:= NullString;

    if phfPrintDate in fPageHeaderOps then
    begin
      HdrLine:= DateToStr( PrintingDateTime );
    end;

    if phfPrintTime in fPageHeaderOps then
    begin
      if phfPrintDate in fPageHeaderOps then HdrLine:= HdrLine + ',';
      HdrLine:= HdrLine + ' ' +TimeToStr( PrintingDateTime );
    end;

    if phfPageNo in fPageHeaderOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPpageNo)+IntToStr( CurrPageNo );
    end;

(*    if phfNumOfPages in fPageHeaderOps then
    begin
    end; { if } *)

{>>A 18-03-1998}
    if phfHorzPieceNo in fPageHeaderOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPieceNo) + IntToStr(Piece);
    end; { if }

    if phfHorzPiecesTotal in fPageHeaderOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPieceOf) + IntToStr(fHorzPieces);
    end; { if }
{<<A 18-03-1998}

    OneLineHt:= 0;
    if trim( HdrLine )<>NullString then OneLineHT:= PrnCV.TextHeight( HdrLine );

    if fPrintPreview then
    begin
      inc( VOffset, PVoffset );
    end; { if }

    if OneLineHT>0 then
    begin
      if phfBreakLines in fPageHeaderOps then
      begin
        inc( OneLineHT, (HLineWid*2)+round(10*fYAspect) );

        PrnCV.MoveTo( round(fBeginLeft*fXAspect), VOffset );
        PrnCV.LineTo( round(fEndRight*fXAspect), VOffset );

        PrnCV.MoveTo( round(fBeginLeft*fXAspect), VOffset+OneLineHT );
        PrnCV.LineTo( round(fEndRight*fXAspect), VOffset+OneLineHT );
      end; { if }

      ARect:= rect( round(fBeginLeft * fXAspect),
                    round(fBeginTop  * fYAspect) + VOffset,
                    round(fEndRight  * fXAspect),
                    round(fBeginTop  * fYAspect) + OneLineHt +VOffset );

      DrawText( PrnCV.Handle,
                StrPCopy(C, HdrLine ), -1,
                ARect,
                      ( DSPwwraps[fPHeaderWordWrap] or
                         DSPALIGNS[fPHeaderAlignment] or
                         DT_NOPREFIX or DT_EXPANDTABS ) );

    end; { if }

    if Piece=1 then
    begin
      LGDpageStart:= LGSpageStart;
{>>C 09-09-1996}
      inc( LGDpageStart, OneLineHt+PrnCV.TextHeight( FKerningCharacter ) );
    end
      else
    begin
      LGDpageStart:= LGSpageStart;
{>>C 09-09-1996}
      inc( LGDpageStart, OneLineHt+PrnCV.TextHeight( FKerningCharacter ) );
    end; { if }

    Result:= True;
  end;

  {--- print page footer -----------------------------------------------------}

  function PrintPageFooter( Piece : integer ): boolean;
  var
    PageFtrWid: integer;
    OneLineHt : integer;
    LineIdx   : integer;
    HdrLine   : string;
    VOffset   : integer;
    ARect     : TRect;

  begin
    Result:= False;

    if Piece=1 then
    begin
      dec( LGSpageEnd, round(10*FYAspect) );
      if assigned( fOnEndPage ) then fOnEndPage( self );
    end; { if }

    VOffset:= LGSpageEnd;

    PageFtrWid:= fEndRight-fBeginLeft;

    PrnCV.Font:= fPFooterFont;
    PrnCV.Font.Color:= clBlack;
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Width:= HLineWid;
    PrnCV.Pen.Color:= clBlack;

    HdrLine:= NullString;

    if phfPrintDate in fPageFooterOps then
    begin
      HdrLine:= DateToStr( PrintingDateTime );
    end;

    if phfPrintTime in fPageFooterOps then
    begin
      if phfPrintDate in fPageFooterOps then HdrLine:= HdrLine + ',';
      HdrLine:= HdrLine + ' ' +TimeToStr( PrintingDateTime );
    end;

    if phfPageNo in fPageFooterOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPpageNo) +IntToStr( CurrPageNo );
    end;

(*    if phfNumOfPages in fPageFooterOps then
    begin
    end; { if } *)

{>>A 18-03-1998}
    if phfHorzPieceNo in fPageFooterOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPieceNo) + IntToStr(Piece);
    end; { if }

    if phfHorzPiecesTotal in fPageFooterOps then
    begin
      HdrLine:= HdrLine + LoadStr(SDSPieceOf) + IntToStr(fHorzPieces);
    end; { if }
{<<A 18-03-1998}

    OneLineHt:= 0;
    if trim( HdrLine )<>NullString then OneLineHT:= PrnCV.TextHeight( HdrLine );

    VOffset:= VOffset-OneLineHT;

    if fPrintPreview then
    begin
      inc( VOffset, PVoffset );
    end; { if }

    if OneLineHT>0 then
    begin
      if phfBreakLines in fPageFooterOps then
      begin
        inc( OneLineHT, (HLineWid*2)+round(10*fYAspect) );

        PrnCV.MoveTo( round(fBeginLeft*fXAspect), VOffset );
        PrnCV.LineTo( round(fEndRight*fXAspect), VOffset );

        PrnCV.MoveTo( round(fBeginLeft*fXAspect), VOffset+OneLineHT );
        PrnCV.LineTo( round(fEndRight*fXAspect), VOffset+OneLineHT );
      end; { if }

      ARect:= rect( round(fBeginLeft * fXAspect),
                    round(fBeginTop  * fYAspect) + VOffset,
                    round(fEndRight  * fXAspect),
                    round(fBeginTop  * fYAspect) + OneLineHt +VOffset );


      DrawText( PrnCV.Handle,
                StrPCopy(C, HdrLine ), -1,
                ARect,
                ( DSPwwraps[fPFooterWordWrap] or
                  DSPALIGNS[fPFooterAlignment] or
                  DT_NOPREFIX or DT_EXPANDTABS ) );

    end; { if }

    if Piece=1 then
    begin
      LGDPageEnd:= LGSpageEnd;
{>>C 09-09-1996}
      dec( LGDpageEnd, OneLineHt+PrnCV.TextHeight(FKerningCharacter) );
    end; { if }

    Result:= True;
  end;

  {--- PrintAlignmentMarkers -------------------------------------------------}

  procedure PrintAlignmentMarker(const X, Y: integer);
  begin
    PrnCV.MoveTo(ScaleX(X-wCrossMarkerWidth), ScaleY(Y));
    PrnCV.LineTo(ScaleX(X+wCrossMarkerWidth), ScaleY(Y));
    PrnCV.MoveTo(ScaleX(X), ScaleY(Y-wCrossMarkerWidth));
    PrnCV.LineTo(ScaleX(X), ScaleY(Y+wCrossMarkerWidth));
  end; { PrintAlignmentMarker }

  procedure PrintAlignmentMarkers(const PieceNo: integer);
  {- here we prints an alignment markers (crosses) to help users to align clipped
     pages }
  begin
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Color:= clBlack;
    PrnCV.Pen.Width:= 1;

    if PieceNo>1 then
    begin {- print left side markers }
      PrintAlignmentMarker(fBeginLeft, fBeginTop+1);
      PrintAlignmentMarker(fBeginLeft, fEndBottom-3);
    end; { if }

    if PieceNo<fHorzPieces then
    begin {- print right side markers }
      PrintAlignmentMarker(fEndRight-2, fBeginTop+1);
      PrintAlignmentMarker(fEndRight-2, fEndBottom-3);
    end; { if }
  end; { PrintAlignmentMarkers }

  {--- PrintPiece ------------------------------------------------------------}

  function PrintPiece( PP : integer ): boolean;
  var
    Idx     : integer;
    CWid    : integer;
    Cleft   : integer;
    TmpHT   : integer;
    ARect   : TRect;

    VOffset : integer;
    HOffset : integer;

    FirstOnPage : boolean;
    ColnVisible : boolean;
    PrintingExtraRows: boolean;
    ExtraRowNo: integer;
  begin
    Result:= False;

    IsLastPage:= False;
    PrintingExtraRows:= FDataSet.EOF;
    ExtraRowNo:= 0;    

    try
      DataSet.GotoBookmark( BKM );
    except
      Screen.Cursor:= crDefault;
      MessageDlg(LoadStr(SDSPgotoBKMerror), mtError, [mbOK], 0 );

      Exit;
    end; { try }

{>>A 12-12-1997}
    if FOverlayPrinting and (fHorzPieces>1) then
      PrintAlignmentMarkers(PP);
{<<A 12-12-1997}

{- Print page header }
    PrintPageHeader( PP );

{- Print page footer }
    PrintPageFooter( PP );

    FirstOnPage:= True;

    VOffset:= LGDpageStart;
    HOffset:= fBeginLeft;

{>>C 08-09-1996 }
    PrnCV.Font:= FDataFont;
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Color:= clBlack;

    if fPrintPreview then
    begin
      HOffset:= ScaleX( HOffset );

      inc( VOffset, PVoffset );
    end; { if }

    PrnCV.Font:= fDataHdrFont;
    PrnCV.Font.Color:= clBlack;
{>>C 09-09-1996}
    GridHdrHt:= PrnCV.TextHeight(FKerningCharacter)+(HLineWid*2);

    for Idx:= 0 to fFieldCells.Count-1 do
    begin
      if (fFieldCells.Objects[Idx] as TFieldCell).UnPrn then
        Continue;

      ColnVisible:= (fFieldCells.Objects[Idx] as TFieldCell).Visible;

      if fPrintVisibleOnly and (not ColnVisible) then
        Continue;

      if (not fPrintMemos) and
         (DataSet.Fields[Idx].DataType in dtsMemos) then
        Continue;

      {- Skip invisible cells if PrintVisibleOnly property is True }

      CWid:=  (fFieldCells.Objects[Idx] as TFieldCell).Width;
      CLeft:= (fFieldCells.Objects[Idx] as TFieldCell).Left;

      if fPrintPreview then
      begin
        CWid:=  ScaleX( CWid );
        CLeft:= ScaleX( CLeft );
      end; { if }

{>>>} if DataSet.Fields[Idx].DataType in dtsMemos then
      begin
        {!!!!}
      end
        else
      begin
        TmpHT:= GetCellHeight( TrimTrail(ReplaceWrappers(DataSet.Fields[Idx].DisplayLabel,
                                                         FDTWThreadingMode )),
           rect( CLeft+round(4*fXAspect)+VLineWid,
                 VOffset+round(4*fYAspect)+HLineWid,
                 CLeft+CWid-round(4*fXAspect),
                 GridHdrHt ),
                 fDataHdrWordWrap,
                 fDataHdrAlignment,
                 PrnCV );
      end; { if else }

{- We're find the maximum cells' height to use }
      if TmpHT>GridHdrHt then GridHdrHt:= TmpHT;
    end; { for }

    inc( GridHdrHt, round(4*fYAspect)+HLineWid*2 );

 {- Printing cells' titles }
    for Idx:= 0 to fFieldCells.Count-1 do
    begin
      if (fFieldCells.Objects[Idx] as TFieldCell).Piece<>PP then
        Continue;
      {- Cell is not from current piece, skip them }

      if (fFieldCells.Objects[Idx] as TFieldCell).UnPrn then
        Continue;
      {- Cell is not to be printed, skip them }

      ColnVisible:= (fFieldCells.Objects[Idx] as TFieldCell).Visible;

      if fPrintVisibleOnly and
         (not ColnVisible) then
        Continue;
      {- Skip invisible cells if PrintVisibleOnly property is True }

      if (not fPrintMemos) and
         (DataSet.Fields[Idx].DataType in dtsMemos) then
        Continue;
      {- Skip memo field if not PrintMemos }

      CWid:=  (fFieldCells.Objects[Idx] as TFieldCell).Width;
      CLeft:= (fFieldCells.Objects[Idx] as TFieldCell).Left;

      if fPrintPreview then
      begin
        CWid:=  integer(round( CWid  * fXAspect ));
        CLeft:= integer(round( CLeft * fXAspect ));
      end; { if }

      inc( CWid, VLineWid );

      PrnCV.Font:= fDataHdrFont;
      PrnCV.Font.Color:= clBlack;

{>>C 1.23}
      {- horz lines }
      if ddoHorzGridLines in FDataOptions then
      begin
        PrnCV.Pen.Width:= HLineWid;

        PrnCV.MoveTo( CLeft, VOffset );
        PrnCV.LineTo( CLeft+CWid, VOffset );
      end; { for }

      {- vert lines }
      if ddoVertGridLines in FDataOptions then
      begin
        PrnCV.Pen.Width:= VLineWid;

        if FirstOnPage then
        begin {- draws leftmost/vert. line }
          PrnCV.MoveTo( CLeft, VOffset );
          PrnCV.LineTo( CLeft, VOffset+GridHdrHt );

          FirstOnPage:= False;
        end; { if }

        {- draws right/vert. line }
        PrnCV.MoveTo( CLeft+CWid, VOffset );
        PrnCV.LineTo( CLeft+CWid, VOffset+GridHdrHt );
      end; { if }
{<<C 1.23}

      if fDataHdrAlignment<>taRightJustify then
      begin
        ARect:= rect( CLeft+round(4*fXAspect)+VLineWid,
                      VOffset+round(4*fYAspect)+HLineWid,
                      CLeft+CWid-round(4*fXAspect),
                      VOffset+GridHdrHt-round(4*fYAspect) );

        DrawText( PrnCV.Handle,
                  StrPCopy(C, TrimTrail(ReplaceWrappers(DataSet.Fields[Idx].DisplayLabel,
                                                        FDTWThreadingMode ))),
                  -1,
                  ARect,
                  ( DSPwwraps[fDataHdrWordWrap] or
                    DSPALIGNS[fDataHdrAlignment] or
                    DT_NOPREFIX or DT_EXPANDTABS ) );
      end
        else
      begin
        ARect:= rect( CLeft+round(4*fXAspect),
                      VOffset+round(4*fYAspect)+HLineWid,
                      CLeft+CWid-round(4*fXAspect)-VLineWid,
                      VOffset+GridHdrHt-round(4*fYAspect) );

        DrawText( PrnCV.Handle,
                  StrPCopy(C, TrimTrail(ReplaceWrappers(DataSet.Fields[Idx].DisplayLabel,
                                                        FDTWThreadingMode ))),
                  -1,
                  ARect,
                  ( DSPwwraps[fDataHdrWordWrap] or
                    DSPALIGNS[fDataHdrAlignment] or
                    DT_NOPREFIX or DT_EXPANDTABS ) );
      end; { if else }
    end; { for }

    VOffset:= VOffset+GridHdrHt;

    if fPrintPreview then
      PVoffset:= PVOffset + GridHdrHt;

{- Now print the data grid }

    while ((not DataSet.EOF) or
      ((PrintingExtraRows) and (ExtraRowNo+1 <= fExtraGridRows.RowCount))) and
      ((VOffset+CHt)<LGDpageEnd) do
    begin
      PrnCV.Font:= fDataFont;
      PrnCV.Font.Color:= clBlack;
{>>C 09-09-1996}
      CHt:= PrnCV.TextHeight(FKerningCharacter)+(HLineWid*2);

      for Idx:= 0 to fFieldCells.Count-1 do
      begin
        if (fFieldCells.Objects[Idx] as TFieldCell).UnPrn then
          Continue;

        ColnVisible:= (fFieldCells.Objects[Idx] as TFieldCell).Visible;

        if fPrintVisibleOnly and (not ColnVisible) then
          Continue;
        {- Skip invisible cells if PrintVisibleOnly property is True }

        if (not fPrintMemos) and
           (DataSet.Fields[Idx].DataType in dtsMemos) then
             Continue;

        CWid:=  (fFieldCells.Objects[Idx] as TFieldCell).Width;
        CLeft:= (fFieldCells.Objects[Idx] as TFieldCell).Left;

        if fPrintPreview then
        begin
          CWid:=  integer(round( CWid  * fXAspect ));
          CLeft:= integer(round( CLeft * fXAspect ));

          PVoffset:= PVOffset + CHt;
        end; { if }

{>>A 1.23: Assigning value addicted to be printed to be used in later ops. }
        if PrintingExtraRows then
          FDataString:= TrimTrail(FExtraGridRows.Cells[Idx, ExtraRowNo])
        else
          FDataString:= TrimTrail(DataSet.Fields[Idx].DisplayText);
{<<A 1.23}

        TmpHT:= GetCellHeight( FDataString,
             rect( CLeft+round(4*fXAspect) + VLineWid,
                   round(4*fYAspect),
                   CLeft+CWid,
                   CHt ),
             fWordWrapCells,
             DataSet.Fields[Idx].Alignment,
             PrnCV );

{- We're find the maximum cells' height to use }
        if TmpHT>CHt then CHt:= TmpHT;
      end; { for }

      inc( CHt, HLineWid*2 );

{- Now prints only these cells which have the Piece number= passed ( PP ) }

      FirstOnPage:= True;

      for Idx:= 0 to fFieldCells.Count-1 do
      begin
{>>C 12-12-1997: added overlayed printing }
        if FOverlayPrinting then
        begin
          if ((fFieldCells.Objects[Idx] as TFieldCell).Piece=(PP-1)) and
             (not (fFieldCells.Objects[Idx] as TFieldCell).Clipped) then
            Continue;
          {- Cell is not an clipped predecessor, skip it }
        end; { if }

        if (fFieldCells.Objects[Idx] as TFieldCell).Piece<>PP then
          Continue;
        {- Cell is not from current piece, skip it }

        if (fFieldCells.Objects[Idx] as TFieldCell).UnPrn then
          Continue;

        ColnVisible:= (fFieldCells.Objects[Idx] as TFieldCell).Visible;

        if fPrintVisibleOnly and (not ColnVisible) then
          Continue;
        {- Skip invisible cells if PrintVisibleOnly property is True }

        if (not fPrintMemos) and
           (DataSet.Fields[Idx].DataType in dtsMemos) then
          Continue;

        CWid:=  (fFieldCells.Objects[Idx] as TFieldCell).Width;

        if ((fFieldCells.Objects[Idx] as TFieldCell).Piece=(PP-1)) and
           ((fFieldCells.Objects[Idx] as TFieldCell).Clipped) then
          CLeft:= (fFieldCells.Objects[Idx] as TFieldCell).Left -
            (fFieldCells.Objects[Idx] as TFieldCell).ClipLeft
        else
          CLeft:= (fFieldCells.Objects[Idx] as TFieldCell).Left;

        if fPrintPreview then
        begin
          CWid:=  integer(round( CWid  * fXAspect ));
          CLeft:= integer(round( CLeft * fXAspect ));

          PVoffset:= PVOffset + CHt;
        end; { if }

        inc( CWid, VLineWid );

{- Printing data cells }
        PrnCV.Font:= fDataFont;
        PrnCV.Font.Color:= clBlack;

{>>C 1.23}
        {- horz lines }
        if ddoHorzGridLines in FDataOptions then
        begin
          PrnCV.Pen.Width:= HLineWid;

          PrnCV.MoveTo( CLeft, VOffset );
          PrnCV.LineTo( CLeft+CWid, VOffset );
        end; { if }

        {- vert lines }
        if ddoVertGridLines in FDataOptions then
        begin
          PrnCV.Pen.Width:= VLineWid;

          if FirstOnPage then
          begin {- draws leftmost vert. line }
            PrnCV.MoveTo( CLeft, VOffset );
            PrnCV.LineTo( CLeft, VOffset+CHt );

            FirstOnPage:= False;
          end; { if }

          {- draws right/vert line }
          PrnCV.MoveTo( CLeft+CWid, VOffset );
          PrnCV.LineTo( CLeft+CWid, VOffset+CHt );
        end; { if }
{<<C 1.23}

{>>A 1.23: Assigning value addicted to be printed to be used in later ops. }
        if PrintingExtraRows then
          FDataString:= TrimTrail(FExtraGridRows.Cells[Idx, ExtraRowNo])
        else
          FDataString:= TrimTrail(DataSet.Fields[Idx].DisplayText);
{<<A 1.23}
{>>F14: 10-12-1997}
        if DataSet.Fields[Idx].Alignment<>taRightJustify then
        begin
          ARect:= rect( CLeft+round(4*fXAspect)+VLineWid,
                        VOffset+round(2*fYAspect)+HLineWid,
                        CLeft+CWid,
                        VOffset+CHt-round(2*fYAspect) );
{<<F14}

          DrawText( PrnCV.Handle,
{>>C 1.23: Printing the FDataString value assigned before }
                    StrPCopy(C, FDataString), -1,
                    ARect,
                    ( DSPwwraps[fWordWrapCells] or
                       DSPALIGNS[DataSet.Fields[Idx].Alignment] or
                       DT_NOPREFIX or DT_EXPANDTABS ) )
{<<C 1.23}
        end
          else
        begin
{>>F14: 10-12-1997}
          ARect:= rect( CLeft+round(4*fXAspect),
                        VOffset+round(2*fYAspect)+HLineWid,
                        CLeft+CWid-VLineWid,
                        VOffset+CHt-round(2*fYAspect) );
{<<F14}

          DrawText( PrnCV.Handle,
                    StrPCopy(C, FDataString), -1,
                    ARect,
                    ( DSPwwraps[fWordWrapCells] or
                       DSPALIGNS[DataSet.Fields[Idx].Alignment] or
                       DT_NOPREFIX or DT_EXPANDTABS ) );
        end; { if else }
      end; { for }

      VOffset:= VOffset+CHt;

      try
        DataSet.Next;
      except
        Screen.Cursor:= crDefault;
        MessageDlg(LoadStr(SDSPnextRecError), mtError, [mbOK], 0 );

        Exit;
      end; { try }

{<<A 1.23}
      if PrintingExtraRows then
        inc(ExtraRowNo);

      { Toggles flag if DataSet is at EOD and there is not a page's end }
      if DataSet.EOF and ((VOffset+CHt)<LGDpageEnd) then
        IsLastPage:= True;

      if IsLastPage then
        PrintingExtraRows:= True;  
{>>A 1.23}
{>>C 1.23}
      if (ddoHorzGridLines in FDataOptions) and
         ((VOffset+CHt>=LGDpageEnd) or (DataSet.EOF)) then
      begin
        PrnCV.Pen.Width:= HLineWid;
        PrnCV.MoveTo( round(fBeginLeft*fXAspect), VOffset );
        PrnCV.LineTo( CLeft+CWid, VOffset );
      end; { if }
{<<C 1.23}

      LGDlastVofs:= VOffset;

      inc( ProgressCnt );
      UpdateNXprogressDlg( NullString, ProgressCnt, Cancelled );

      if Cancelled then
        Break;
    end; { while }

    if fPrintPreview then
    begin
      inc( PVoffset, VOffset + 10 );
    end; { if }

    Result:= True;
  end; { PrintPiece }

{-----------------------------------------------------------------------------}

  procedure BeginDocument;
  begin
    if not fPrintPreview and not Assigned(FPrinterCanvas) then
    begin
      Printer.BeginDoc;
      Printer.Title:= Application.Title+' '+LoadStr(SDSPjobHdrInc);
    end; { if }
  end; { BeginDocument }

  procedure EndDocument;
  begin
    if not fPrintPreview and not Assigned(FPrinterCanvas) then
    begin
      Printer.EndDoc;
    end; { if }
  end; { EndDocument }

  {--- print report title ----------------------------------------------------}

  function PrintTitle : boolean;
  var
    TtlCellWid: integer;
    OneLineHt : integer;
    LineIdx   : integer;
    VOffset   : integer;
    ARect     : TRect;

  begin
    Result:= False;

    VOffset:=  round(10*FYAspect);
    TtlCellWid:= round((fEndRight*FXAspect)-(fBeginLeft*FXAspect));

    PrnCV.Font:= fTitleFont;
    PrnCV.Font.Color:= clBlack;
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Width:= FHorzLineWidth;
    PrnCV.Pen.Color:= clBlack;

    for LineIdx:= 0 to fTitle.Count-1 do
    begin
      ARect:= rect( round(fBeginLeft*fXAspect) + VOffset,
                    round(fBeginTop*fYAspect),
                    round(fEndRight*fXAspect),
                    round((fBeginTop+10)*fYAspect) + VOffset );

      OneLineHt:=
        DrawText( PrnCV.Handle,
                  StrPCopy(C, fTitle[LineIdx] ), -1,
                  ARect,
                  (  DT_CALCRECT or
                     DSPwwraps[fTitleWordWrap] or
                     DSPALIGNS[fTitleAlignment] or
                     DT_NOPREFIX or DT_EXPANDTABS ) );

      if OneLineHt<0 then
      begin
        Screen.Cursor:= crDefault;
        MessageDlg(LoadStr(SDSPNegDTretError), mtError, [mbCancel], 0 );

        Break;
      end; { if }

      ARect:= rect( round(fBeginLeft * fXAspect),
                    round(fBeginTop  * fYAspect) +VOffset,
                    round(fEndRight  * fXAspect),
                    round(fBeginTop  * fYAspect) + OneLineHt +VOffset );

      DrawText( PrnCV.Handle,
                StrPCopy(C, fTitle[LineIdx] ), -1,
                ARect,
                ( DSPwwraps[fTitleWordWrap] or
                  DSPALIGNS[fTitleAlignment] or
                  DT_NOPREFIX or DT_EXPANDTABS ) );

      inc( VOffset, OneLineHt );
    end; { for }

    inc( LGSpageStart, VOffset+round(10*fYAspect) );

    Result:= True;
  end; { PrintTitle }

  {--- print report tail -----------------------------------------------------}

  function PrintTail : boolean;
  var
    TailCellWid: integer;
    OneLineHt : integer;
    LineIdx   : integer;
    VOffset   : integer;
    ARect     : TRect;

  begin
    if fTail.Count<=0 then
    begin {- exiting if Tail contains no data }
      Result:= True;

      Exit;
    end; { if }

    Result:= False;

{>>C 09-09-1996}
    VOffset:= LGDlastVOfs+PrnCV.TextHeight( FKerningCharacter );
    TailCellWid:= fEndRight-fBeginLeft;

    if fHorzPieces>1 then
    begin
      BeginPage;

      LGSpageStart:= fBeginTop;

      PrintPageHeader( 1 );
      PrintPageFooter( 1 );

      VOffset:= LGDpageStart;
    end; { if }

    PrnCV.Font:= fTailFont;
    PrnCV.Font.Color:= clBlack;
    PrnCV.Brush.Style:= bsClear;
    PrnCV.Pen.Width:= FHorzLineWidth;
    PrnCV.Pen.Color:= clBlack;

    for LineIdx:= 0 to fTail.Count-1 do
    begin
      ARect:= rect( round(fBeginLeft * fXAspect),
                    round(fBeginTop  * fYAspect),
                    round(fEndRight  * fXAspect),
                    round((fBeginTop+10)*fYAspect) );

      OneLineHt:=
        DrawText( PrnCV.Handle,
                  StrPCopy(C, fTail[LineIdx] ), -1,
                  ARect,
                  ( DSPwwraps[fTailWordWrap] or
                    DT_CALCRECT or
                    DSPALIGNS[fTailAlignment] or
                    DT_NOPREFIX or DT_EXPANDTABS ) );

      if OneLineHt<0 then
      begin
        Screen.Cursor:= crDefault;
        MessageDlg(LoadStr(SDSPNegDTretError), mtError, [mbCancel], 0 );

        Break;
      end; { if }

      if VOffset+OneLineHT>LGSpageEnd then
      begin
        if fPrintPreview then Break;

        BeginPage;

        inc( CurrPageNo );

        PrintPageHeader( 1 );
        PrintPageFooter( 1 );

        VOffset:= LGDpageStart;
      end; { if }

      ARect:= rect( round(fBeginLeft * fXAspect),
                    round(fBeginTop  * fYAspect) +VOffset,
                    round(fEndRight  * fXAspect),
                    round(fBeginTop  * fYAspect) + OneLineHt +VOffset );

      DrawText( PrnCV.Handle,
                StrPCopy(C, fTail[LineIdx] ), -1,
                ARect,
                ( DSPwwraps[fTailWordWrap] or
                  DSPALIGNS[fTailAlignment] or
                  DT_NOPREFIX or DT_EXPANDTABS ) );

      inc( VOffset, OneLineHt );
    end; { for }

    Result:= True;
  end; { PrintTail }

{--- generic print method ----------------------------------------------------}

begin { GenericPrint }
  Result:= False;

  Screen.Cursor:= crHourGlass;

  if not ReadyToPrint then
  begin
    Screen.Cursor:= crDefault;

    Exit;
  end; { if }

  if not CalculatePrinterProps then
  begin
    Screen.Cursor:= crDefault;

    Exit;
  end; { if }

  if not BuildFieldCells then
  begin
    ClearFieldCells;

    Screen.Cursor:= crDefault;

    Exit;
  end; { if }

  try
    DStateBKM:= DataSet.GetBookMark;
  except
    Screen.Cursor:= crDefault;
    MessageDlg(LoadStr(SDSPPutBKMerror), mtError, [mbCancel], 0 );

    Exit;
  end; { try }

  DataSet.DisableControls;

  if fRestartData then
  try
    DataSet.First;
  except
    Screen.Cursor:= crDefault;
    MessageDlg(LoadStr(SDSPBODerror), mtError, [mbCancel], 0 );

    DataSet.FreeBookmark( DStateBKM );
    DataSet.EnableControls;

    Exit;
  end; { try }

  {- Now begin printing }

  if Assigned(PrinterCanvas) then
  begin
    PrnCV := PrinterCanvas;
  end
  else
  begin
    if fPrintPreview then
    begin
      PrnCV:= PPform.PPV.Canvas;
    end
    else
    begin
      PrnCV:= Printer.Canvas;
    end; { if }
  end; { if }

  {--- Beginning document ----------------------------------------------------}

  LGSpageStart:= fBeginTop;
  LGSpageEnd  := fEndBottom;

  CurrPageNo:= 1;
  TotalPages:= 0;

  BeginDocument;

{-- triggering an event BeforePrint }
    if assigned( fOnBeforePrint ) then
      fOnBeforePrint( self );

  {--- Printing report header ------------------------------------------------}

  PVOffset:= 0;
  PrnCV.Font:= fTitleFont;

  if not PrintTitle then
  begin
    Screen.Cursor:= crDefault;
    MessageDlg(LoadStr(SDSPhdrPrintError), mtError, [mbOK], 0 );
    EndDocument;

    DataSet.FreeBookmark( DStateBKM );
    DataSet.EnableControls;

    Exit;
  end; { if }

  {--- Printing data grid ----------------------------------------------------}

  PrnCV.Font:= fDataFont;

{>>C 09-09-1996}
  CHt:= Printer.Canvas.TextHeight(FKerningCharacter);

  HasErrors:= False;

  ProgressCnt:= 0;
  Cancelled:= False;

  try
    if not fPrintPreview then
      if not CreateNXProgressDlg( LoadStr(SDSPprogressHdr),
                                  LoadStr(SDSPprogressMsg),
                                  0,
                                  DataSet.RecordCount*longint(fHorzPieces), 0 )
      then
      begin
      end; { if }

    while not DataSet.EOF do
    begin
      BKM:= DataSet.GetBookmark;

{>>C 12-12-1997: control piece's # to print}
      if PieceNo=0 then
      begin
        for CurrPiece:=1 to fHorzPieces do
        begin
          HasErrors:= not PrintPiece( CurrPiece );

          if HasErrors then Break;

          if Cancelled then Break;

          if CurrPiece<fHorzPieces then BeginPage;
        end; { for }
      end
      else
      begin
{>>A 12-12-1997}
        CurrPiece:= PieceNo;
        HasErrors:= not PrintPiece( CurrPiece );

        if HasErrors then Break;

        if Cancelled then Break;
{<<A 12-12-1997}
      end; { if }
{<<C 12-12-1997: control piece's # to print}

      LGSpageStart:= fBeginTop;
      LGSpageEnd:= fEndBottom;

      inc( CurrPageNo );

      DataSet.FreeBookMark( BKM );

      if Cancelled then Break;

      if DataSet.EOF then
      begin
        Break;
      end; { if }

      if fPrintPreview then
      begin
        Break;
      end; { if }

      if HasErrors then
      begin
        Break;
      end; { if }

      try
        DataSet.Next;
      except
      end; { try }

      inc( ProgressCnt );
      if not fPrintPreview then
        UpdateNXprogressDlg( NullString, ProgressCnt, Cancelled );

      if Cancelled then Break;

      if not DataSet.EOF then BeginPage;
    end; { while }

{-- triggering an event AfterPrint }
    if assigned( fOnAfterPrint ) then fOnAfterPrint( self );

    PrintTail;

{--- Ending document -------------------------------------------------------}

    EndDocument;

  except
    on E : Exception do
    begin
      MessageDlg(E.Message, mtError, [mbOK], 0 );
    end;
  end;

  Screen.Cursor:= crDefault;

  DestroyNXprogressDlg;

  try
    DataSet.GotoBookmark( DStateBKM );
  except
    Screen.Cursor:= crDefault;
    MessageDlg(LoadStr(SDSPgotoBKMerror), mtError, [mbOK], 0 );
  end; { try }

  DataSet.FreeBookmark( DStateBKM );
  DataSet.EnableControls;

  Result:= True;
end;

{-----------------------------------------------------------------------------}

Function  TDataSetPrint.Print : Boolean;
begin
  Result:= False;

  fPrintPreview:= False;

{>>A 18-03-1998}
  FPrintingDateTime:= Now;
{<<A 18-03-1998}

  Result:= GenericPrint(0);

  ClearFieldCells;
end;

{/////////////////////////////////////////////////////////////////////////////}

Function TDataSetPrint.Preview : Boolean;
begin
  Result:= False;

  PrintPreview:= True;

  if fPrintPreview then
  try
    PPform:= TPrintPreviewForm.Create( Application );
  except
    MessageDlg(LoadStr(SDSPpreviewFrmError), mtError, [mbCancel], 0 );

    ClearFieldCells;

    raise;

    Exit;
  end; { try }

  try
    PPForm.DSP:= self;

    CalculatePrinterProps;

    PPForm.PPV.Width:=  ScaleX( Printer.PageWidth );
    PPForm.PPV.Height:= ScaleY( Printer.PageHeight );

{>>A 18-03-1998}
  FPrintingDateTime:= Now;
{<<A 18-03-1998}

    PPForm.ShowModal;
  except
    MessageDlg(LoadStr(SDSPpreviewFrmError), mtError, [mbCancel], 0 );
  end; { try }

  PPForm.Free;

  ClearFieldCells;
end;

{-----------------------------------------------------------------------------}

{/// Service functions ///////////////////////////////////////////////////////}

function  TDataSetPrint.MM2PixelsVert( MMs : integer ) : integer;
begin
  if MMs=0 then begin Result:= 0; Exit; end;
  Result:= integer( round( (MMs / MMperInch) * DevPPI_Y ) );
end;

function  TDataSetPrint.MM2PixelsHorz( MMs : integer ) : integer;
begin
  if MMs=0 then begin Result:= 0; Exit; end;
  Result:= integer( round( (MMs / MMperInch) * DevPPI_X ) );
end;

{/////////////////////////////////////////////////////////////////////////////}

end.



